@@ -1,22 +1,30 @@
 package POE::Filter::IRCD;
+{
+  $POE::Filter::IRCD::VERSION = '2.44';
+}
+
+#ABSTRACT: A POE-based parser for the IRC protocol
 
 use strict;
+use warnings;
 use Carp;
-use vars qw($VERSION);
-use base qw(POE::Filter);
-
-$VERSION = '2.30';
+use base qw[POE::Filter];
 
 sub _PUT_LITERAL () { 1 }
 
 # Probably some other stuff should go here.
 
-our $g = {
+my $g = {
   space			=> qr/\x20+/o,
   trailing_space	=> qr/\x20*/o,
 };
 
-our $irc_regex = qr/^
+my $irc_regex = qr/^
+  (?:
+    \x40                # '@'-prefixed IRCv3.2 messsage tags.
+    (\S+)               # [tags] Semi-colon delimited key=value list
+    $g->{space}
+  )?
   (?:
     \x3a                #  : comes before hand
     (\S+)               #  [prefix]
@@ -36,7 +44,7 @@ our $irc_regex = qr/^
         $g->{'space'}
         [^\x00\x0a\x0d\x20\x3a]
         [^\x00\x0a\x0d\x20]*
-      ){0,13}           # then match on 0-13 of these,
+      )*                # then match as many of these as possible
     )
   )?                    # otherwise dont match at all.
   (?:
@@ -72,15 +80,21 @@ sub get {
 
   foreach my $raw_line (@$raw_lines) {
     warn "->$raw_line \n" if $self->{DEBUG};
-    if ( my($prefix, $command, $middles, $trailing) = $raw_line =~ m/$irc_regex/ ) {
+    if ( my($tags, $prefix, $command, $middles, $trailing) = $raw_line =~ m/$irc_regex/ ) {
       my $event = { raw_line => $raw_line };
+      if ($tags) {
+        for my $tag_pair (split /;/, $tags) {
+          my ($thistag, $thisval) = split /=/, $tag_pair;
+          $event->{tags}->{$thistag} = $thisval
+        }
+      }
       $event->{'prefix'} = $prefix if $prefix;
       $event->{'command'} = uc $command;
       $event->{'params'} = [] if defined ( $middles ) || defined ( $trailing );
       push @{$event->{'params'}}, (split /$g->{'space'}/, $middles) if defined $middles;
       push @{$event->{'params'}}, $trailing if defined $trailing;
       push @$events, $event;
-    } 
+    }
     else {
       warn "Received line $raw_line that is not IRC protocol\n";
     }
@@ -99,15 +113,21 @@ sub get_one {
 
   if ( my $raw_line = shift ( @{ $self->{BUFFER} } ) ) {
     warn "->$raw_line \n" if $self->{DEBUG};
-    if ( my($prefix, $command, $middles, $trailing) = $raw_line =~ m/$irc_regex/ ) {
+    if ( my($tags, $prefix, $command, $middles, $trailing) = $raw_line =~ m/$irc_regex/ ) {
       my $event = { raw_line => $raw_line };
+      if ($tags) {
+        for my $tag_pair (split /;/, $tags) {
+          my ($thistag, $thisval) = split /=/, $tag_pair;
+          $event->{tags}->{$thistag} = $thisval
+        }
+      }
       $event->{'prefix'} = $prefix if $prefix;
       $event->{'command'} = uc $command;
       $event->{'params'} = [] if defined ( $middles ) || defined ( $trailing );
       push @{$event->{'params'}}, (split /$g->{'space'}/, $middles) if defined $middles;
       push @{$event->{'params'}}, $trailing if defined $trailing;
       push @$events, $event;
-    } 
+    }
     else {
       warn "Received line $raw_line that is not IRC protocol\n";
     }
@@ -128,6 +148,15 @@ sub put {
       my $colonify = ( defined $event->{colonify} ? $event->{colonify} : $self->{COLONIFY} );
       if ( _PUT_LITERAL || _checkargs($event) ) {
         my $raw_line = '';
+        if ( ref $event->{tags} eq 'HASH' && keys %{ $event->{tags} } ) {
+          $raw_line .= '@';
+          my @tags = %{ $event->{tags} };
+          while (my ($thistag, $thisval) = splice @tags, 0, 2) {
+            $raw_line .= $thistag . ( defined $thisval ? '='.$thisval : '' );
+            $raw_line .= ';' if @tags;
+          }
+          $raw_line .= ' ';
+        }
         $raw_line .= (':' . $event->{'prefix'} . ' ') if exists $event->{'prefix'};
         $raw_line .= $event->{'command'};
 	if ( $event->{'params'} and ref $event->{'params'} eq 'ARRAY' ) {
@@ -143,11 +172,11 @@ sub put {
 	}
         push @$raw_lines, $raw_line;
         warn "<-$raw_line \n" if $self->{DEBUG};
-      } 
+      }
       else {
         next;
       }
-    } 
+    }
     else {
       warn __PACKAGE__ . " non hashref passed to put(): \"$event\"\n";
       push @$raw_lines, $event if ref $event eq 'SCALAR';
@@ -156,6 +185,13 @@ sub put {
   return $raw_lines;
 }
 
+sub clone {
+  my $self = shift;
+  my $nself = { };
+  $nself->{$_} = $self->{$_} for keys %{ $self };
+  $nself->{BUFFER} = [ ];
+  return bless $nself, ref $self;
+}
 
 # This thing is far from correct, dont use it.
 sub _checkargs {
@@ -178,9 +214,15 @@ sub _checkargs {
 
 __END__
 
+=pod
+
 =head1 NAME
 
-POE::Filter::IRCD -- A POE-based parser for the IRC protocol.
+POE::Filter::IRCD - A POE-based parser for the IRC protocol
+
+=head1 VERSION
+
+version 2.44
 
 =head1 SYNOPSIS
 
@@ -206,9 +248,9 @@ A standalone version exists as L<Parse::IRC>.
 
 =over
 
-=item new
+=item C<new>
 
-Creates a new POE::Filter::IRCD object. Takes two optional arguments: 
+Creates a new POE::Filter::IRCD object. Takes two optional arguments:
 
   'debug', which will print all lines received to STDERR;
   'colonify', set to 1 to force the filter to always colonify the last param passed in a put(),
@@ -220,13 +262,13 @@ Creates a new POE::Filter::IRCD object. Takes two optional arguments:
 
 =over
 
-=item get_one_start
+=item C<get_one_start>
 
-=item get_one
+=item C<get_one>
 
-=item get_pending
+=item C<get_pending>
 
-=item get
+=item C<get>
 
 Takes an arrayref which is contains lines of IRC formatted input. Returns an arrayref of hashrefs
 which represents the lines. The hashref contains the following fields:
@@ -234,7 +276,7 @@ which represents the lines. The hashref contains the following fields:
   prefix
   command
   params ( this is an arrayref )
-  raw_line 
+  raw_line
 
 For example, if the filter receives the following line, the following hashref is produced:
 
@@ -247,35 +289,31 @@ For example, if the filter receives the following line, the following hashref is
 		raw_line => ':moo.server.net 001 lamebot :Welcome to the IRC network lamebot',
 	   }
 
-=item put
+=item C<put>
 
 Takes an arrayref containing hashrefs of IRC data and returns an arrayref containing IRC formatted lines.
 Optionally, one can specify 'colonify' to override the global colonification option.
 eg.
 
-  $hashref = { 
-		command => 'PRIVMSG', 
-		prefix => 'FooBar!foobar@foobar.com', 
+  $hashref = {
+		command => 'PRIVMSG',
+		prefix => 'FooBar!foobar@foobar.com',
 		params => [ '#foobar', 'boo!' ],
 		colonify => 1, # Override the global colonify option for this record only.
 	      };
 
   $filter->put( [ $hashref ] );
 
-=item debug
+=item C<clone>
+
+Makes a copy of the filter, and clears the copy's buffer.
+
+=item C<debug>
 
 With a true or false argument, enables or disables debug output respectively. Without an argument the behaviour is to toggle the debug status.
 
 =back
 
-=head1 MAINTAINER
-
-Chris Williams <chris@bingosnet.co.uk>
-
-=head1 AUTHOR
-
-Jonathan Steinert
-
 =head1 SEE ALSO
 
 L<POE>
@@ -290,5 +328,15 @@ L<POE::Component::IRC>
 
 L<Parse::IRC>
 
-=cut
+=head1 AUTHOR
 
+Chris Williams <chris@bingosnet.co.uk>
+
+=head1 COPYRIGHT AND LICENSE
+
+This software is copyright (c) 2013 by Chris Williams and Jonathan Steinert.
+
+This is free software; you can redistribute it and/or modify it under
+the same terms as the Perl 5 programming language system itself.
+
+=cut
